"""A generator used to generate C++ based on a schema."""

from logging import Logger
from pathlib import Path
import jinja2
from cmg.templates import (
    cmakelists_txt_j2,
    klass_cpp_j2,
    test_cpp_j2,
    identifiable_hpp_j2,
    identifiable_cpp_j2,
    index_hpp_j2,
    index_cpp_j2,
    persistable_hpp_j2,
    persistence_hpp_j2,
)
import cmg.templates.klass_hpp_j2 as klass_hpp_j2
from cmg.schema import Schema, Klass, Field
from importlib.machinery import SourceFileLoader

from cmg.validation import SchemaRuleSet


def schema_loader(schema: str) -> Schema:
    """Load the schema module."""

    module = SourceFileLoader("schema", schema).load_module()
    return module.schema


def generate(schema: Schema, output_dir: str, logger: Logger) -> int:
    """Generate C++ code based on the schema."""

    logger.info("Validating schema ...")
    errors = SchemaRuleSet().validate(schema)

    if len(errors) > 0:
        for error in errors:
            logger.error(error.message)
        return 1

    logger.info("Generating code ...")

    # Create the directory
    Path(output_dir).mkdir(parents=True, exist_ok=True)

    # Link the schema
    schema.link()
    schema.set_output_dir(output_dir)

    # Build the C++ code

    # Templates
    hpp_template = jinja2.Template(klass_hpp_j2.TEMPLATE)
    cpp_template = jinja2.Template(klass_cpp_j2.TEMPLATE)
    supplementary_templates = {
        "identifiable.hpp": jinja2.Template(identifiable_hpp_j2.TEMPLATE),
        "identifiable.cpp": jinja2.Template(identifiable_cpp_j2.TEMPLATE),
        "index.hpp": jinja2.Template(index_hpp_j2.TEMPLATE),
        "index.cpp": jinja2.Template(index_cpp_j2.TEMPLATE),
        "persistable.hpp": jinja2.Template(persistable_hpp_j2.TEMPLATE),
        "persistence.hpp": jinja2.Template(persistence_hpp_j2.TEMPLATE),
    }
    cmakelists_template = jinja2.Template(cmakelists_txt_j2.TEMPLATE)
    test_template = jinja2.Template(test_cpp_j2.TEMPLATE)

    for klass in schema.classes:
        hpp_file = f"{output_dir}/{klass.to_snake_case()}.hpp"
        with open(hpp_file, "w") as f:
            f.write(hpp_template.render(schema=schema, klass=klass))
        cpp_file = f"{output_dir}/{klass.to_snake_case()}.cpp"
        with open(cpp_file, "w") as f:
            f.write(cpp_template.render(schema=schema, klass=klass))

    supplementary_classes = [
        "identifiable",
        "index",
    ]
    for klass in supplementary_classes:
        header_file = f"{output_dir}/{klass}.hpp"
        with open(header_file, "w") as f:
            f.write(supplementary_templates[f"{klass}.hpp"].render(schema=schema))
        cpp_file = f"{output_dir}/{klass}.cpp"
        with open(cpp_file, "w") as f:
            f.write(supplementary_templates[f"{klass}.cpp"].render(schema=schema))

    header_only_classes = ["persistable", "persistence"]
    for klass in header_only_classes:
        header_file = f"{output_dir}/{klass}.hpp"
        with open(header_file, "w") as f:
            f.write(supplementary_templates[f"{klass}.hpp"].render(schema=schema))

    cmakelists_file = f"{output_dir}/CMakeLists.txt"
    with open(cmakelists_file, "w") as f:
        f.write(cmakelists_template.render(schema=schema))

    test_file = f"{output_dir}/test_{schema.namespace}.cpp"
    with open(test_file, "w") as f:
        f.write(test_template.render(schema=schema))

    return 0



TEMPLATE = """
#ifndef _IDENTIFIABLE_HPP_
#define _IDENTIFIABLE_HPP_

#include <memory>

namespace {{schema.namespace}} {
    /**
     * @brief Identifiable class, inherit from this class to make an object identifiable in indecies and persistent storage
     */

    class Index;

    class Identifiable
    {
    public:
        Identifiable() = default;
        virtual ~Identifiable() = default;

        /**
         * @brief Get the id of the object
         * @return The id of the object
         */
        unsigned long getId() const;

        /**
         * @brief Set the id of the object
         * @param id The id of the object
         */
        void setId(unsigned long id);

        /**
         * @brief Add the object to an index
         */
        void addToIndex(std::shared_ptr<Index> index);

    protected:
        unsigned long id = 0;

    };

}
#endif
"""



TEMPLATE = """
#include <stdexcept>
#include "{{klass.to_snake_case()}}.hpp"
{%- for include in klass.get_forward_includes() %}
#include "{{include}}.hpp"
{%- endfor %}

namespace {{schema.namespace}}
{

    /*
    unsigned long {{klass.name}}::getId() const {
        return Identifiable::getId();
    }

    void {{klass.name}}::setId(unsigned long id) {
        Identifiable::setId(id);
    }
    */
    
    {{klass.get_create_ptr_type()}} {{klass.name}}::create({{klass.get_create_arguments()}})
    {
            auto object = std::make_shared<{{klass.name}}>(PrivateConstructor());
{%- for field in klass.init_fields(parents=False) %}
            object->{{field.to_camel_case()}} = {{field.to_camel_case()}};
{%- endfor %}
{%- for field in klass.fields %}
    {%- if field.has_parent() %}
            object->{{field.to_camel_case()}} = {{field.to_camel_case()}} ;
        {%- if field._parent_field.is_list %}
            object->{{field.to_camel_case()}}.lock()->addTo{{field._parent_field.to_camel_case(upper_first=True)}}(object, true);
        {%- else %}
            object->{{field.to_camel_case()}}.lock()->set{{field._parent_field.to_camel_case(upper_first=True)}}(object);
        {%- endif %}
    {%- endif %}
{%- endfor %}
            return object;
    }

    void {{klass.name}}::update(std::map<std::string, std::any> fields)
    {
        for (const auto &[key, value] : fields)
        {
{%- for field in klass.get_updatable_fields() %}
            if (key == "{{field.to_camel_case()}}")
            {
                set{{field.to_camel_case(upper_first=True)}}(std::any_cast<{{field.get_cpp_type(cast=True)}}>(value));
                continue;
            }
{%- endfor %}
            throw std::invalid_argument("Invalid field: " + key);
        }
    }

    void {{klass.name}}::destroy(bool fromParent)
    {
{%- for field in klass.get_ordered_fields() %}
    {%- if field.has_parent() %}

        // Remove from parent field for {{field.to_camel_case()}}
        if (!fromParent && !{{field.to_camel_case()}}.expired())
        {
        {%- if field._parent_field.is_list %}
            {{field.to_camel_case()}}.lock()->removeFrom{{field._parent_field.to_camel_case(upper_first=True)}}(getptr());
        {%- else %}
            {{field.to_camel_case()}}.lock()->set{{field._parent_field.to_camel_case(upper_first=True)}}(nullptr);
        {%- endif %}
        }
    {%- endif %}
{%- endfor %}

{%- for field in klass.get_ordered_fields() %}
    {%- if field.is_child %}

        // Destory child field(s) for {{field.to_camel_case()}}
        {%- if field.is_list %}
        std::vector<std::shared_ptr<{{field._child_klass.name}}>> toDestroy = {};
        for (auto &item : {{field.to_camel_case()}})
        {
            toDestroy.push_back(item);
        }
        for (auto &item : toDestroy)
        {
            item->destroy(true);
        }
        {{field.to_camel_case()}}.clear();
        {{field.to_camel_case()}}.shrink_to_fit();
        {%- else %}
        if ({{field.to_camel_case()}} != nullptr)
        {
            {{field.to_camel_case()}}->destroy(true);
        }
        {{field.to_camel_case()}} = nullptr;
        {%- endif %}
    {%- endif %}
{%- endfor %}
    }

    std::shared_ptr<{{klass.name}}> {{klass.name}}::getptr()
    {
        return shared_from_this();
    }
    
{%- for field in klass.get_ordered_fields() %}
    {%- if field.has_parent() %}
        {% if field.is_list %}
    throw std::runtime_error("Cannot set parent field for list for class {{klass.name}} and field {{field.to_camel_case()}}"); 
        {% else %}
    void {{klass.name}}::set{{field.to_camel_case(upper_first=True)}}({{field.get_cpp_type()}} parent)
    {
        {%- if field._parent_field.is_list %}
        // Remove from current parent
        if (!{{field.to_camel_case()}}.expired())
        {
            {{field.to_camel_case()}}.lock()->removeFrom{{field._parent_field.to_camel_case(upper_first=True)}}(getptr());
            {{field.to_camel_case()}}.reset();
        }

        // Add to new parent
        if (!parent.expired())
        {
            parent.lock()->addTo{{field._parent_field.to_camel_case(upper_first=True)}}(getptr(), true);
            {{field.to_camel_case()}} = parent;
        }
        {%- else %}
        // Remove from current parent
        if (!{{field.to_camel_case()}}.expired())
        {
            {{field.to_camel_case()}}.lock()->set{{field._parent_field.to_camel_case(upper_first=True)}}(nullptr);
            {{field.to_camel_case()}}.reset();
        }

        // Add to new parent
        if (!parent.expired())
        {
            parent.lock()->set{{field._parent_field.to_camel_case(upper_first=True)}}(getptr());
            {{field.to_camel_case()}} = parent;
        }
        {%- endif %}
    }
        {% endif %}
    {%- else %}
        {% if field.is_list %}
            {%- if field._child_field %}
    void {{klass.name}}::addTo{{field.to_camel_case(upper_first=True)}}({{field.get_cpp_type(nolist=True)}} item, bool fromChild)
    {
        if (fromChild)
        {
            {{field.to_camel_case()}}.push_back(item);
        }
        else
        {
            item->set{{field._child_field.to_camel_case(upper_first=True)}}(getptr());
        }
    
    }

    void {{klass.name}}::removeFrom{{field.to_camel_case(upper_first=True)}}({{field.get_cpp_type(nolist=True)}} item, bool fromChild)
    {
        if (fromChild)
        {
            {{field.to_camel_case()}}.erase(std::remove({{field.to_camel_case()}}.begin(), {{field.to_camel_case()}}.end(), item), {{field.to_camel_case()}}.end());
        }
        else
        {
            item->reset{{field._child_field.to_camel_case(upper_first=True)}}();
        }
    
    }
            {%- else %}

    void {{klass.name}}::addTo{{field.to_camel_case(upper_first=True)}}({{field.get_cpp_type(nolist=True)}} item)
    {
        {{field.to_camel_case()}}.push_back(item);
    }

    void {{klass.name}}::removeFrom{{field.to_camel_case(upper_first=True)}}({{field.get_cpp_type(nolist=True)}} item)
    {
        {{field.to_camel_case()}}.erase(std::remove({{field.to_camel_case()}}.begin(), {{field.to_camel_case()}}.end(), item), {{field.to_camel_case()}}.end());
    }
            {%- endif %}
        {% else %}
    void {{klass.name}}::set{{field.to_camel_case(upper_first=True)}}({{field.get_cpp_type()}} {{field.to_camel_case()}})
    {
        this->{{field.to_camel_case()}} = {{field.to_camel_case()}};
    }

        {% endif %}
    {%- endif %}

    {%- if field.has_parent(): %}
    void  {{klass.name}}::reset{{field.to_camel_case(upper_first=True)}}()
    {
        if (!{{field.to_camel_case()}}.expired())
        {
    {%- if field._parent_field.is_list %}
            {{field.to_camel_case()}}.lock()->removeFrom{{field._parent_field.to_camel_case(upper_first=True)}}(getptr(), true);
    {%- else %}
            {{field.to_camel_case()}}.lock()->set{{field._parent_field.to_camel_case(upper_first=True)}}(nullptr);
    {%- endif %}
            {{field.to_camel_case()}}.reset();
        }
    }
    {%- endif %}

    {{field.get_cpp_type()}} {{klass.name}}::get{{field.to_camel_case(upper_first=True)}}()
    {
        return {{field.to_camel_case()}};
    }


{%- endfor %}

    void {{klass.name}}::addToIndex(std::shared_ptr<Index> index)
    {
        index->add(getptr());
{%- for field in klass.get_ordered_fields() %}
    {%- if field.is_child %}

        // Add {{field.to_camel_case()}} to index
        {%- if field.is_list %}
        for (auto &item : {{field.to_camel_case()}})
        {
            item->addToIndex(index);
        }
        {%- else %}
        if ({{field.to_camel_case()}} != nullptr)
        {
            {{field.to_camel_case()}}->addToIndex(index);
        }
        {%- endif %}
    {%- endif %}
{%- endfor %}
    }

    void {{klass.name}}::serializeFields(std::ostream &os)
    {
        // Serialize ID
        os.write(reinterpret_cast<const char *>(&id), sizeof(id));

{%- for field in klass.get_ordered_fields() %}

    {%- if field.name != "id"  %}
        {%- if field.has_parent() %}

        // Serialize {{field.to_camel_case()}}
        unsigned long {{field.to_camel_case()}}Id = {{field.to_camel_case()}}.lock() ? {{field.to_camel_case()}}.lock()->getId() : 0;
        os.write(reinterpret_cast<const char *>(&{{field.to_camel_case()}}Id), sizeof({{field.to_camel_case()}}Id));
        {%- elif field.is_child %}
            {%- if field.is_list %}

        // Serialize {{field.to_camel_case()}}
        size_t {{field.to_camel_case()}}Size = {{field.to_camel_case()}}.size();
        os.write(reinterpret_cast<const char *>(&{{field.to_camel_case()}}Size), sizeof({{field.to_camel_case()}}Size));
        for (const auto &item : {{field.to_camel_case()}})
        {
            item->serializeFields(os);
        }
            {%- else %}

        // Serialize {{field.to_camel_case()}}
        if ({{field.to_camel_case()}} == nullptr)
        {
            unsigned long nullId = 0;
            os.write(reinterpret_cast<const char *>(&nullId), sizeof(nullId));
        }
        else
        {
            unsigned long itemId = {{field.to_camel_case()}}->getId();
            os.write(reinterpret_cast<const char *>(&itemId), sizeof(itemId));
            {{field.to_camel_case()}}->serializeFields(os);
        }
            {%- endif %}
        {%- else %}
            {%- if field.get_cpp_type() == "std::string" %}

        // Serialize {{field.to_camel_case()}}
        size_t {{field.to_camel_case()}}Size = {{field.to_camel_case()}}.size();
        os.write(reinterpret_cast<const char *>(&{{field.to_camel_case()}}Size), sizeof({{field.to_camel_case()}}Size));
        os.write({{field.to_camel_case()}}.c_str(), {{field.to_camel_case()}}.size());
            {%- elif not field.is_reference() %}

        // Serialize {{field.to_camel_case()}}
        os.write(reinterpret_cast<const char *>(&{{field.to_camel_case()}}), sizeof({{field.to_camel_case()}}));
            {%- endif %}
        {%- endif %}
    {%- endif %}
{%- endfor %}
    }

    void {{klass.name}}::serializeReferences(std::ostream &os)
    {
{%- for field in klass.get_ordered_fields() %}
    {%- if field.is_reference() and not field.has_parent() and not field.is_child %}
        // Serialize {{field.to_camel_case()}}
        if ( {{field.to_camel_case()}}.expired() )
        {
            unsigned long nullId = 0;
            os.write(reinterpret_cast<const char *>(&nullId), sizeof(nullId));
        }
        else
        {
            unsigned long {{field.to_camel_case()}}Id = {{field.to_camel_case()}}.lock()->getId();
            os.write(reinterpret_cast<const char *>(&{{field.to_camel_case()}}Id), sizeof({{field.to_camel_case()}}Id));
        }
    {%- elif field.is_child %}
        // Serialize {{field.to_camel_case()}}
        {%- if field.is_list %}
        for (const auto &item : {{field.to_camel_case()}})
        {
            item->serializeReferences(os);
        }
        {%- else %}
        if ({{field.to_camel_case()}} != nullptr)
        {
            {{field.to_camel_case()}}->serializeReferences(os);
        }
        {%- endif %}
    {%- endif %}
{%- endfor %}
    }

    std::shared_ptr<{{klass.name}}> {{klass.name}}::deserializeFields(std::istream &is, std::shared_ptr<Index> index)
    {
        auto object = std::make_shared<{{klass.name}}>(PrivateConstructor());
        
        // Deserialize ID
        is.read(reinterpret_cast<char *>(&object->id), sizeof(object->id));

        // Add to index
        object->addToIndex(index);

{%- for field in klass.get_ordered_fields() %}
    {%- if field.name != "id"  %}
        {%- if field.has_parent() %}
        
        // Deserialize {{field.to_camel_case()}}
        unsigned long {{field.to_camel_case()}}Id;
        is.read(reinterpret_cast<char *>(&{{field.to_camel_case()}}Id), sizeof({{field.to_camel_case()}}Id));
        object->{{field.to_camel_case()}} = std::dynamic_pointer_cast<{{field.type}}>(index->get({{field.to_camel_case()}}Id));

        {%- elif field.is_child %}
            {%- if field.is_list %}

        // Deserialize {{field.to_camel_case()}}
        size_t {{field.to_camel_case()}}Size;
        is.read(reinterpret_cast<char *>(&{{field.to_camel_case()}}Size), sizeof({{field.to_camel_case()}}Size));
        for (size_t i = 0; i < {{field.to_camel_case()}}Size; i++)
        {
            auto item = {{field._child_klass.name}}::deserializeFields(is, index);
            object->{{field.to_camel_case()}}.push_back(item);
        }
            {%- else %}

        // Deserialize {{field.to_camel_case()}}
        unsigned long {{field.to_camel_case()}}Id;
        is.read(reinterpret_cast<char *>(&{{field.to_camel_case()}}Id), sizeof({{field.to_camel_case()}}Id));
        if ({{field.to_camel_case()}}Id != 0)
        {
            object->{{field.to_camel_case()}} = {{field._child_klass.name}}::deserializeFields(is, index);
        }
            {%- endif %}
        {%- else %}
            {%- if field.get_cpp_type() == "std::string" %}

        // Deserialize {{field.to_camel_case()}}
        size_t {{field.to_camel_case()}}Size;
        is.read(reinterpret_cast<char *>(&{{field.to_camel_case()}}Size), sizeof({{field.to_camel_case()}}Size));
        char *{{field.to_camel_case()}}Buffer = new char[{{field.to_camel_case()}}Size];
        is.read({{field.to_camel_case()}}Buffer, {{field.to_camel_case()}}Size);
        object->{{field.to_camel_case()}} = std::string({{field.to_camel_case()}}Buffer, {{field.to_camel_case()}}Size);
        delete[] {{field.to_camel_case()}}Buffer;
            {%- elif not field.is_reference() %}

        // Deserialize {{field.to_camel_case()}}
        is.read(reinterpret_cast<char *>(&object->{{field.to_camel_case()}}), sizeof(object->{{field.to_camel_case()}}));
            {%- endif %}
        {%- endif %}
    {%- endif %}
{%- endfor %}

        return object;
    }

    void {{klass.name}}::deserializeReferences(std::istream &is, std::shared_ptr<Index> index)
    {
{%- for field in klass.get_ordered_fields() %}
    {%- if field.is_reference() and not field.has_parent() and not field.is_child %}

        // Deserialize {{field.to_camel_case()}}
        unsigned long {{field.to_camel_case()}}Id;
        is.read(reinterpret_cast<char *>(&{{field.to_camel_case()}}Id), sizeof({{field.to_camel_case()}}Id));
        if ({{field.to_camel_case()}}Id != 0)
        {
            {{field.to_camel_case()}} = std::dynamic_pointer_cast<{{field.type}}>(index->get({{field.to_camel_case()}}Id));
        }
    {%- elif field.is_child %}

        // Deserialize {{field.to_camel_case()}}
        {%- if field.is_list %}
        for (auto &item : {{field.to_camel_case()}})
        {
            item->deserializeReferences(is, index);
        }
        {%- else %}
        if ({{field.to_camel_case()}} != nullptr)
        {
            {{field.to_camel_case()}}->deserializeReferences(is, index);
        }
        {%- endif %}
    {%- endif %}
{%- endfor %}
    }
}
"""



TEMPLATE = """
#ifndef _INDEX_HPP_
#define _INDEX_HPP_

#include <memory>
#include <map>

#include "identifiable.hpp"

namespace {{schema.namespace}}
{
    /**
     * @brief Index class, used to store objects by id
     */

    class Index
    {
        public:
            Index() = default;
            ~Index() = default;


            /**
                * @brief Add an object to the index
                * @param object The object to add
            */      
            void add(std::shared_ptr<Identifiable> object);

            /**
                * @brief Remove an object from the index
                * @param object The object to remove
            */
            void remove(std::shared_ptr<Identifiable> object);

            /**
                * @brief Get an object by id
                * @param id The id of the object to get
                * @return The object with the given id
            */
            std::shared_ptr<Identifiable> get(unsigned long id);

        private:
            unsigned long nextId = 1;
            std::map<unsigned long, std::shared_ptr<Identifiable>> objects;

    };
}
#endif
"""



TEMPLATE = """
cmake_minimum_required(VERSION 3.10)

# Set the project name
project({{schema.namespace}})

# Specify the C++ standard
set(CMAKE_CXX_STANDARD 17)
set(CMAKE_CXX_STANDARD_REQUIRED True)
set(CMAKE_CXX_FLAGS "${CMAKE_CXX_FLAGS} -Wall -Werror -fprofile-instr-generate -fcoverage-mapping")
cmake_policy(SET CMP0135 NEW)

add_library({{schema.namespace}} "identifiable.cpp" "index.cpp" {{schema.get_cmakelists_src()}} )

# Testing
include(FetchContent)

FetchContent_Declare(
  googletest
  URL https://github.com/google/googletest/archive/03597a01ee50ed33e9dfd640b249b4be3799d395.zip
)
# For Windows: Prevent overriding the parent project's compiler/linker settings
set(gtest_force_shared_crt ON CACHE BOOL "" FORCE)
FetchContent_MakeAvailable(googletest)

enable_testing()

add_executable(
  test_{{schema.namespace}}
  test_{{schema.namespace}}.cpp
)
target_link_libraries(
  test_{{schema.namespace}}
  GTest::gtest_main
  {{schema.namespace}}
)


include(GoogleTest)
gtest_discover_tests(test_{{schema.namespace}})

"""



import click
import logging

from cmg import generator

"""
CLI for the cmg package.

Example usage:

cmg --schema <path to schema file> --output <path to output directory>
"""


@click.command()
@click.option("--schema", required=True, help="Path to schema file")
@click.option("--output", required=True, help="Path to output directory")
def cli(schema, output):
    """
    Generate code from a schema file.
    """
    logging.basicConfig(
        level=logging.INFO,
        format="%(levelname)s: %(message)s",
    )
    logger: logging.Logger = logging.getLogger("cmg")
    logger.info(f"Generating code from schema {schema} to output directory {output}")
    exit_code = generator.generate(generator.schema_loader(schema), output, logger)
    if exit_code != 0:
        logger.error("Code generation failed.")
    else:
        logger.info("Code generation complete.")

    exit(exit_code)


if __name__ == "__main__":
    cli()



TEMPLATE = """

#ifndef _PERSITANCE_HPP_
#define _PERSITANCE_HPP_

#include <memory>
#include <fstream>
#include <string>
#include <stdexcept>
#include "index.hpp"
#include "persistable.hpp"
#include "identifiable.hpp"

namespace solar_system
{
    /**
     * @brief Persistence class, used to save and load objects
     */
    template <typename T, typename = std::enable_if_t<std::is_base_of_v<Persistable<T>, T> && std::is_base_of_v<Identifiable, T>>>
    class Persistence
    {
    public:
        Persistence() {
            index = std::make_shared<Index>();
        };

        ~Persistence() = default;

        /*
            @brief Save the root object to a file
            @param root The root object to save
            @param filename The filename to save to
        */
        void save(std::shared_ptr<T> root, std::string filename)
        {
            // Add root object to index
            root->addToIndex(index);

            // Save root object
            std::ofstream os(filename, std::ios::binary);

            // Write schema version
            std::string schemaVersion = "{{schema.version}}";
            size_t versionSize = schemaVersion.size();
            os.write(reinterpret_cast<const char *>(&versionSize), sizeof(size_t));
            os.write(schemaVersion.c_str(), schemaVersion.size());

            // Serialize
            root->serializeFields(os);
            root->serializeReferences(os);

            // Close file
            os.close();
        }

        /**
         * @brief Load a root object from a file
         * @param filename The filename to load from
         */
        std::shared_ptr<T> load(std::string filename)
        {
            // Open file
            std::ifstream is(filename, std::ios::binary);

            // Read schema version
            size_t versionSize;
            is.read(reinterpret_cast<char *>(&versionSize), sizeof(size_t));
            char *versionBuffer = new char[versionSize];
            is.read(versionBuffer, versionSize);
            std::string schemaVersion(versionBuffer, versionSize);
            delete[] versionBuffer;

            // Check schema version
            if (schemaVersion != "{{schema.version}}")
            {
                throw std::runtime_error("Invalid schema version");
            }

            // Create index
            auto index = std::make_shared<Index>();

            // Deserialize
            auto root = T::deserializeFields(is, index);
            root->addToIndex(index);
            root->deserializeReferences(is, index);

            // Close file
            is.close();

            return root;
        }

    private:
        std::shared_ptr<Index> index;
    };
}

#endif
"""



TEMPLATE = """
#include <gtest/gtest.h>

{%- for include in schema.get_test_includes() %}
{{include}}
{%- endfor %}
#include "index.hpp"
#include "identifiable.hpp"
#include "persistence.hpp"

using namespace {{schema.namespace}};

{%- for klass in schema.classes %}

class Test{{klass.name}} : public ::testing::Test
{
protected:
    void SetUp() override
    {
    {%- for field in klass.get_parent_fields() %}
        // Create parent object
        {{field.to_camel_case()}} = {{field._parent_klass.name}}::create({{field._parent_klass.get_example_arguments()}}){% if loop.index > 1 %}.lock()->getptr(){% endif %};
    {%- endfor %}
        // Create test object
    {%- if klass.has_parent() %}
        {{klass.get_var_name()}} = {{klass.name}}::create({{klass.get_example_arguments()}}).lock()->getptr();
    {%- else %}
        {{klass.get_var_name()}} = {{klass.name}}::create({{klass.get_example_arguments()}});
    {%- endif %}
    {%- for field in klass.get_ordered_fields() %}
        {%- if field.is_child %}
        auto child{{loop.index}} = {{field._child_klass.name}}::create({{field._child_klass.get_example_arguments()}}).lock()->getptr();
            {%- if field.is_list %}
        {{klass.get_var_name()}}->addTo{{field.to_camel_case(upper_first=True)}}(child{{loop.index}});
            {%- else %}
        {{klass.get_var_name()}}->set{{field.to_camel_case(upper_first=True)}}(child{{loop.index}});
            {%- endif %}
        {%- endif %}    
    {%- endfor %}


    }

    {%- for field in klass.get_parent_fields() %}
    std::shared_ptr<{{field._parent_klass.name}}> {{field.to_camel_case()}};
    {%- endfor %}
    std::shared_ptr<{{klass.name}}> {{klass.get_var_name()}};
};

{%- endfor %}

{%- for klass in schema.classes %}

TEST_F(Test{{klass.name}}, test_id)
{
    // Check object id
    EXPECT_EQ({{klass.get_var_name()}}->getId(), 0);
    {{klass.get_var_name()}}->setId(1);
    EXPECT_EQ({{klass.get_var_name()}}->getId(), 1);
}

TEST_F(Test{{klass.name}}, test_create)
{
    // Check object creation
    EXPECT_NE({{klass.get_var_name()}}, nullptr);
}

TEST_F(Test{{klass.name}}, test_update_exception)
{
    // Check exception thrown when updating with invalid field
    EXPECT_THROW(
        {
        try
        {
            {{klass.get_var_name()}}->update({%raw%}{{"missing", std::string("missing")}}{%endraw%});
        }
        catch (const std::invalid_argument &e)
        {
            // and this tests that it has the correct message
            EXPECT_STREQ("Invalid field: missing", e.what());
            throw;
        } }, std::invalid_argument);
}

TEST_F(Test{{klass.name}}, test_update)
{
    // Check object update
    {{klass.get_var_name()}}->update({{klass.get_example_update_arguments()}});
    {%- for field in klass.get_updatable_fields() %}
        {%- if field.get_example() != "nullptr" %}
            {%- if field.has_parent() %}
    EXPECT_EQ({{klass.get_var_name()}}->get{{field.to_camel_case(upper_first=True)}}().lock(), {{field.get_example()}});
            {%- else %}
    EXPECT_EQ({{klass.get_var_name()}}->get{{field.to_camel_case(upper_first=True)}}(), {{field.get_example()}});
            {%- endif %}
        {%- endif %}
    {%- endfor %}

    {%- if klass.has_parent() %}
    
    // Test parent field
        {%- for field in klass.get_parent_fields() %}
    auto other{{field.to_camel_case(upper_first=True)}} = {{field._parent_klass.name}}::create({{field._parent_klass.get_example_arguments()}}){% if loop.index > 1 %}.lock()->getptr(){% endif %};
        {%- endfor %}
    {%- set parent_field = klass.get_parent_field() %}
    {%- set parent_field_name = parent_field.to_camel_case() %}
    {%- set parent_field_name_title = parent_field.to_camel_case(upper_first=True) %}
    {{klass.get_var_name()}}->update({%raw%}{{{%endraw%}"{{parent_field_name}}", other{{parent_field_name_title}}{%raw%}}}{%endraw%});
    EXPECT_EQ({{klass.get_var_name()}}->get{{parent_field_name_title}}().lock(), other{{parent_field_name_title}});
        {%- if parent_field._parent_field.is_list %}
    EXPECT_EQ(other{{parent_field_name_title}}->get{{parent_field._parent_field.to_camel_case(upper_first=True)}}().size(), 1);
    EXPECT_EQ({{parent_field_name}}->get{{parent_field._parent_field.to_camel_case(upper_first=True)}}().size(), 0);
        {%- else %}
    EXPECT_EQ(other{{parent_field_name_title}}->get{{parent_field._parent_field.to_camel_case(upper_first=True)}}(), {{klass.get_var_name()}});
    EXPECT_EQ({{parent_field_name}}->get{{parent_field._parent_field.to_camel_case(upper_first=True)}}(), nullptr);
        {%- endif %}
    {%- endif %}

}

    {%- for field in klass.get_updatable_fields() %}
        {% if field.get_example() != "nullptr" and not field.has_parent() %}
TEST_F(Test{{klass.name}}, test_{{field.to_camel_case()}})
{
    // Check object update
    {{klass.get_var_name()}}->set{{field.to_camel_case(upper_first=True)}}({{field.get_example()}});
    EXPECT_EQ({{klass.get_var_name()}}->get{{field.to_camel_case(upper_first=True)}}(), {{field.get_example()}});
}
        {% endif %}
    {%- endfor %}

TEST_F(Test{{klass.name}}, test_destroy)
{
    // Check object destruction
    {{klass.get_var_name()}}->destroy();
}

TEST_F(Test{{klass.name}}, test_save_load)
{
    // Check object save and load
    auto persistance = solar_system::Persistence<{{schema.namespace}}::{{klass.name}}>();
    {{klass.get_var_name()}}->update({{klass.get_example_update_arguments()}});
    persistance.save({{klass.get_var_name()}}, "{{klass.get_var_name()}}.db");
    auto {{klass.get_var_name()}}Loaded = persistance.load("{{klass.get_var_name()}}.db");

    {%- for field in klass.get_ordered_fields() %}
        {%- if field.get_example() != "nullptr" %}
            {%- if field.is_child %}
                {%- if field.is_list %}
    EXPECT_EQ({{klass.get_var_name()}}Loaded->get{{field.to_camel_case(upper_first=True)}}().size(), {{klass.get_var_name()}}->get{{field.to_camel_case(upper_first=True)}}().size());
                {%- else %}
    EXPECT_EQ({{klass.get_var_name()}}Loaded->get{{field.to_camel_case(upper_first=True)}}().getId(), {{klass.get_var_name()}}->get{{field.to_camel_case(upper_first=True)}}().getId());
                {%- endif %}
            {%- elif not field.has_parent() %}
    EXPECT_EQ({{klass.get_var_name()}}Loaded->get{{field.to_camel_case(upper_first=True)}}(), {{field.get_example()}});
            {%- endif %}
        {%- endif %}
    {%- endfor %}

    {%- for field in klass.get_ordered_fields() %}
        {%- if field.is_child and field.is_list %}
    // Create 100 new child objects
    for (int i = 0; i < 100; i++)
    {
        {{field._child_klass.name}}::create({{field._child_klass.get_example_arguments()}});
    }

    persistance.save({{klass.get_var_name()}}, "{{klass.get_var_name()}}_plus100.db");
        {%- endif %}
    {%- endfor %}

}

{%- endfor %}

class TestIndex : public ::testing::Test {

    protected:
        void SetUp() override
        {
            // Create test objects
            object1 = std::make_shared<Identifiable>();
            object2 = std::make_shared<Identifiable>();
            object3 = std::make_shared<Identifiable>();

            // Create index
            index = Index();
        }
    public:
        std::shared_ptr<Identifiable> object1;
        std::shared_ptr<Identifiable> object2;
        std::shared_ptr<Identifiable> object3;
        Index index;
};

TEST_F(TestIndex, test_add_remove_get)
{
    // Check object addition
    index.add(object1);
    EXPECT_EQ(index.get(1), object1);
    index.add(object2);
    EXPECT_EQ(index.get(2), object2);
    index.add(object3);
    EXPECT_EQ(index.get(3), object3);

    // Check object removal
    index.remove(object2);
    EXPECT_EQ(index.get(1), object1);
    EXPECT_EQ(index.get(2), nullptr);
    EXPECT_EQ(index.get(3), object3);

    // Check object addition after removal
    object2->setId(0);
    index.add(object2);
    EXPECT_EQ(index.get(1), object1);
    EXPECT_EQ(index.get(2), nullptr);
    EXPECT_EQ(index.get(3), object3);
    EXPECT_EQ(index.get(4), object2);
}

TEST_F(TestIndex, test_add_existing)
{
    // Check object with exisiting ID not in index already
    object1->setId(99);
    index.add(object1);
    EXPECT_EQ(index.get(99), object1);
    index.add(object2);
    EXPECT_EQ(index.get(100), object2);
}

TEST_F(TestIndex, test_add_existing_in_index)
{
    // Check object with exisiting ID already in index
    index.add(object1);
    EXPECT_EQ(index.get(1), object1);
    index.add(object1);
    EXPECT_EQ(index.get(1), object1);
}

TEST_F(TestIndex, test_add_exisiting_mismatched)
{
    // Check object with exisiting ID already in index but different object
    index.add(object1);
    EXPECT_EQ(index.get(1), object1);
    object2->setId(1);
    EXPECT_THROW(index.add(object2), std::invalid_argument);
}


#include <cstdlib>
class ShellCommandEnvironment : public ::testing::Environment {
public:
    ~ShellCommandEnvironment() override {
        // Run shell commands after all tests
        system("echo All tests finished");
        system("llvm-profdata merge -o default.profdata default.profraw");
        system("llvm-cov export --format=lcov --instr-profile default.profdata test_{{schema.namespace}}.exe --sources {{schema.get_lcov_src()}} > lcov.info");
        system("llvm-cov report --instr-profile default.profdata --ignore-filename-regex='.*_deps.*' test_{{schema.namespace}}.exe --sources {{schema.get_lcov_src()}}");
    }
};

int main(int argc, char** argv) {
    ::testing::InitGoogleTest(&argc, argv);
    ::testing::AddGlobalTestEnvironment(new ShellCommandEnvironment);
    return RUN_ALL_TESTS();
}
"""



TEMPLATE = """
#include <stdexcept>
#include "identifiable.hpp"
#include "index.hpp"

namespace {{schema.namespace}} {
    unsigned long Identifiable::getId() const
    {
        return id;
    }

    void Identifiable::setId(unsigned long id)
    {
        this->id = id;
    }

    void addToIndex(std::shared_ptr<Index> index)
    {
        throw std::runtime_error("Not implemented");
    }
}
"""



"""Validates a schema against a set of rules"""

from abc import ABC
from dataclasses import dataclass
from typing import List, Sequence

from cmg.schema import TYPEMAP, Schema


@dataclass
class RuleError:
    """
    Represents a rule error

    Attributes:
        message: The error message
    """

    message: str


@dataclass
class BaseRule(ABC):
    """
    Base class for all rules

    Attributes:
        name: The name of the rule
        description: The description of the rule
    """

    name: str
    description: str

    def validate(self, schema: Schema) -> List[RuleError]:
        """
        Validates a schema against a rule

        Args:
            schema: The schema to validate

        Returns:
            A list of errors
        """

        raise NotImplementedError


class RuleSet:
    """
    Represents a set of rules

    Attributes:
        rules: The list of rules
    """

    def __init__(self, rules: Sequence[BaseRule]):
        self.rules = rules

    def validate(self, schema) -> List[RuleError]:
        """
        Validates a schema against a set of rules

        Args:
            schema: The schema to validate

        Returns:
            A list of errors
        """

        errors = []

        for rule in self.rules:
            errors += rule.validate(schema)

        return errors


class FieldsHaveValidTypeRules(BaseRule):
    """
    Rule that checks if all fields have a valid type

    Attributes:
        name: The name of the rule
        description: The description of the rule
    """

    def validate(self, schema) -> List[RuleError]:
        # Valid primitive types
        valid_types = list(TYPEMAP.keys())

        # Valid class types
        valid_types.extend([klass.name for klass in schema.classes])

        errors = []
        for klass in schema.classes:
            for field in klass.fields:
                if field.type in valid_types:
                    continue
                errors.append(
                    RuleError(
                        f"Field {field.name} in klass {klass.name} has an invalid type {field.type}"
                    )
                )
        return errors


class FieldExampleMatchesTypeRule(BaseRule):
    """
    Rule that checks if the example matches the field type

    Attributes:
        name: The name of the rule
        description: The description of the rule
    """

    def validate(self, schema) -> List[RuleError]:
        errors = []
        for klass in schema.classes:
            for field in klass.fields:
                has_error = False
                if field.type in TYPEMAP:
                    if not isinstance(field.example, TYPEMAP[field.type][1]):
                        has_error = True

                if has_error:
                    errors.append(
                        RuleError(
                            f"Field {field.name} in klass {klass.name} has an example {field.example} that does not match the type {field.type}"
                        )
                    )
        return errors


class FieldDefaultMatchesTypeRule(BaseRule):
    """
    Rule that checks if the default value matches the field type

    Attributes:
        name: The name of the rule
        description: The description of the rule
    """

    def validate(self, schema) -> List[RuleError]:
        errors = []
        for klass in schema.classes:
            for field in klass.fields:
                if field.default is None:
                    continue
                has_error = False
                if field.type in TYPEMAP:
                    if not isinstance(field.default, TYPEMAP[field.type][1]):
                        has_error = True

                if has_error:
                    errors.append(
                        RuleError(
                            f"Field {field.name} in klass {klass.name} has a default value {field.default} that does not match the type {field.type}"
                        )
                    )
        return errors


class ParentFieldExistsRule(BaseRule):
    """
    Rule that checks if the parent field exists

    Attributes:
        name: The name of the rule
        description: The description of the rule
    """

    def validate(self, schema) -> List[RuleError]:
        errors = []
        for klass in schema.classes:
            for field in klass.fields:
                if field.parent:
                    parent_klass = next(
                        (k for k in schema.classes if k.name == field.type), None
                    )
                    if parent_klass is None:
                        continue

                    parent_field = next(
                        (f for f in parent_klass.fields if f.name == field.parent), None
                    )

                    if parent_field is None:
                        errors.append(
                            RuleError(
                                f"Field {field.name} in klass {klass.name} has a parent field {field.parent} that does not exist in klass {field.type}"
                            )
                        )

        return errors


class KlassNamesUniqueRule(BaseRule):
    """
    Rule that checks if all class names are unique

    Attributes:
        name: The name of the rule
        description: The description of the rule
    """

    def validate(self, schema) -> List[RuleError]:
        errors = []
        class_names = [klass.name for klass in schema.classes]
        visited_klass_names = set()
        for klass in schema.classes:
            if (
                klass.name not in visited_klass_names
                and class_names.count(klass.name) > 1
            ):
                errors.append(RuleError(f"Klass {klass.name} is not unique"))
            visited_klass_names.add(klass.name)
        return errors


class FieldNamesUniqueRule(BaseRule):
    """
    Rule that checks if all field names are unique

    Attributes:
        name: The name of the rule
        description: The description of the rule
    """

    def validate(self, schema) -> List[RuleError]:
        errors = []
        for klass in schema.classes:
            field_names = [field.name for field in klass.fields]
            visited_field_names = set()
            for field in klass.fields:
                if (
                    field.name not in visited_field_names
                    and field_names.count(field.name) > 1
                ):
                    errors.append(
                        RuleError(
                            f"Field {field.name} in klass {klass.name} is not unique"
                        )
                    )
                visited_field_names.add(field.name)
        return errors


class ChildHasParentRule(BaseRule):
    """
    Rule that checks if the child has a parent

    Attributes:
        name: The name of the rule
        description: The description of the rule
    """

    def validate(self, schema) -> List[RuleError]:
        errors = []
        for klass in schema.classes:
            for field in klass.fields:
                if field.is_child:
                    child_klass = next(
                        (k for k in schema.classes if k.name == field.type), None
                    )
                    if child_klass is None:
                        continue

                    parent_field = next(
                        (
                            f
                            for f in child_klass.fields
                            if f.parent == field.name and f.type == klass.name
                        ),
                        None,
                    )

                    if parent_field is None:
                        errors.append(
                            RuleError(
                                f"Field {field.name} in klass {klass.name} is a child field but does not have a parent field in klass {field.type}"
                            )
                        )
        return errors


class SchemaContainsOneRootKlassRule(BaseRule):
    """
    Rule that checks if the schema contains one root klass

    Attributes:
        name: The name of the rule
        description: The description of the rule
    """

    def validate(self, schema) -> List[RuleError]:
        errors = []
        root_klasses = [klass for klass in schema.classes if not klass.has_parent()]
        if len(root_klasses) != 1:
            root_class_names = [klass.name for klass in root_klasses]
            errors.append(
                RuleError(
                    f"Schema must contain exactly one root klass, {len(root_klasses)} found: {', '.join(root_class_names)}"
                )
            )
        return errors


class SchemaRuleSet(RuleSet):
    """
    Rule set for schema validation

    Attributes:
        rules: The list of rules
    """

    def __init__(self):
        rules = [
            FieldsHaveValidTypeRules(
                "FieldsHaveValidTypeRules",
                "Checks if all fields have a valid type",
            ),
            FieldExampleMatchesTypeRule(
                "FieldExampleMatchesTypeRule",
                "Checks if the example matches the field type",
            ),
            ParentFieldExistsRule(
                "ParentFieldExistsRule",
                "Checks if the parent field exists",
            ),
            FieldDefaultMatchesTypeRule(
                "FieldDefaultMatchesTypeRule",
                "Checks if the default value matches the field type",
            ),
            KlassNamesUniqueRule(
                "KlassNamesUniqueRule",
                "Checks if all klass names are unique",
            ),
            FieldNamesUniqueRule(
                "FieldNamesUniqueRule",
                "Checks if all field names are unique",
            ),
            ChildHasParentRule(
                "ChildHasParentRule",
                "Checks if the child has a parent",
            ),
            SchemaContainsOneRootKlassRule(
                "SchemaContainsOneRootKlassRule",
                "Checks if the schema contains one root klass",
            ),
        ]

        super().__init__(rules)



TEMPLATE = """
#include <stdexcept>
#include <string>
#include "index.hpp"

namespace {{schema.namespace}}
{
    void Index::add(std::shared_ptr<Identifiable> object)
    {
        auto objectId = object->getId();

        // If object has no ID, assign next available ID
        if (objectId == 0)
        {
            object->setId(nextId++);
            objects[object->getId()] = object;
        }
        // If object has an ID but is not in the index, add it to the index
        else if (objects.find(objectId) == objects.end())
        {
            // If object ID is greater than next available ID, update next available ID
            if (objectId >= nextId)
            {
                nextId = objectId + 1;
            }

            objects[object->getId()] = object;
        }
        // If object in index is not the same as the object being added, throw an error
        else if (objects[objectId] != object)
        {
            throw std::invalid_argument("Object with id " + std::to_string(objectId) + " already exists in index");
        }
    }

    void Index::remove(std::shared_ptr<Identifiable> object)
    {
        objects.erase(object->getId());
    }

    std::shared_ptr<Identifiable> Index::get(unsigned long id)
    {
        return objects[id];
    }
}
"""



TEMPLATE = """
#ifndef _PERISTABLE_HPP_
#define _PERISTABLE_HPP_

#include <fstream>
#include "index.hpp"

namespace solar_system
{
    /**
     * @brief Persistable class, inherit from this class to make an object persistable from/to a stream
     */
    template <typename T>
    class Persistable
    {
    public:
        Persistable() = default;
        virtual ~Persistable() = default;

        /**
         * @brief Serialize the object's fields and children to a stream
         * @param os The output stream to write to
         */
        virtual void serializeFields(std::ostream &os) = 0;

        /**
         * @brief Serialize the object's references to a stream
         * @param os The output stream to write to
         */
        virtual void serializeReferences(std::ostream &os) = 0;

        /**
         * @brief Deserialize the object's fields and children from a stream
         * @param is The input stream to read from
         * @param index The index to use for reference resolution
         * @return A shared pointer to the deserialized object
         */
        static std::shared_ptr<T> deserializeFields(std::istream &is, std::shared_ptr<Index> index);

        /**
         * @brief Deserialize the object's references from a stream
         * @param is The input stream to read from
         * @param index The index to use for reference resolution
         */
        virtual void deserializeReferences(std::istream &is, std::shared_ptr<Index> index) = 0;
    };
}

#endif
"""



"""A schema used to define the structure of the data in a model."""

from dataclasses import dataclass, field
import os
from typing import Any, List, Optional

"""
Define the type map for C++ types: https://en.cppreference.com/w/cpp/language/types
The key is the type in the schema, the value is a tuple with the C++ type and the Python type
"""
TYPEMAP = {
    "str": ("std::string", str),
    "signed char*": ("std::string", str),
    "unsigned char*": ("std::string", str),
    "char*": ("std::string", str),
    "signed char": ("signed char", int),
    "unsigned char": ("unsigned char", int),
    "char": ("char", int),
    "short": ("short", int),
    "short int": ("short int", int),
    "signed short": ("signed short", int),
    "signed short int": ("signed short int", int),
    "unsigned short": ("unsigned short", int),
    "unsigned short int": ("unsigned short int", int),
    "int": ("int", int),
    "signed": ("signed", int),
    "signed int": ("signed int", int),
    "unsigned": ("unsigned", int),
    "unsigned int": ("unsigned int", int),
    "long": ("long", int),
    "long int": ("long int", int),
    "signed long": ("signed long", int),
    "signed long int": ("signed long int", int),
    "unsigned long": ("unsigned long", int),
    "unsigned long int": ("unsigned long int", int),
    "long long": ("long long", int),
    "long long int": ("long long int", int),
    "signed long long": ("signed long long", int),
    "signed long long int": ("signed long long int", int),
    "unsigned long long": ("unsigned long long", int),
    "unsigned long long int": ("unsigned long long int", int),
    "float": ("float", float),
    "double": ("double", float),
    "long double": ("long double", float),
    "bool": ("bool", bool),
}


def to_snake_case(name: str) -> str:
    """
    Convert a camelCase or PascalCase string to snake_case.

    Args:
        name (str): The name to convert.

    Returns:
        str: The name in snake_case.
    """
    name = "".join(["_" + c.lower() if c.isupper() else c for c in name]).lstrip("_")
    return name


def to_camel_case(name: str | None, upper_first=False) -> str | None:
    """
    Convert a snake_case string to camelCase.

    Args:
        name (str): The name to convert.
        upper_first (bool): Whether to capitalize the first letter.

    Returns:
        str: The name in camelCase.
    """

    if name is None:
        return None
    name = name.replace("_", " ").title().replace(" ", "")
    if upper_first:
        return name
    return name[0].lower() + name[1:]


@dataclass
class Schema:
    """
    A schema used to define the structure of the data in a model.

    Attributes:
        name (str): The name of the schema.

        description (str): The description of the schema.

        namespace (str): The namespace of the schema.

        classes (List[Klass]): The classes in the schema.
    """

    name: str
    description: str
    namespace: str
    version: str
    classes: List["Klass"] = field(default_factory=list)
    _output_dir: str = field(default=".", repr=False, init=False)

    def set_output_dir(self, output_dir: str) -> None:
        """
        Set the output directory for the schema.

        Args:
            output_dir (str): The output directory.
        """
        self._output_dir = os.path.abspath(output_dir)

    def link(self) -> None:
        """
        Link the schema to its classes and fields.
        """
        for klass in self.classes:
            klass._schema = self
            klass.link()

    def get_klass(self, name: str) -> "Klass":
        """
        Get a klass by name.

        Args:
            name (str): The name of the klass.

        Returns:
            Klass: The klass.

        Raises:
            ValueError: If the klass is not found.
        """
        for klass in self.classes:
            if klass.name == name:
                return klass
        raise ValueError(f"Klass not found: {name}")

    def get_field(self, klass_name: str, field_name: str) -> "Field":
        """
        Get a field by klass and field name.

        Args:
            klass_name (str): The name of the klass.
            field_name (str): The name of the field.

        Returns:
            Field: The field.

        Raises:
            ValueError: If the field is not found.
        """
        klass = self.get_klass(klass_name)
        for field in klass.fields:
            if field.name == field_name:
                return field
        raise ValueError(f"Field not found: {field_name} in klass: {klass_name}")

    def get_cmakelists_src(self) -> str:
        """
        Get the source files for the CMakeLists.txt file.

        Returns:
            str: The source files.
        """
        return " ".join([f'"{klass.to_snake_case()}.cpp"' for klass in self.classes])

    def get_lcov_src(self) -> str:
        """
        Get the source files for the lcov command.

        Returns:
            str: The source files.
        """
        sources = [
            "/".join(
                [
                    self._output_dir.replace(os.path.sep, "/").replace("c:", "", 1),
                    f"{klass.to_snake_case()}.cpp",
                ]
            )
            for klass in self.classes
        ]

        sources.extend(
            [
                "/".join(
                    [
                        self._output_dir.replace(os.path.sep, "/").replace("c:", "", 1),
                        f"{klass}.cpp",
                    ]
                )
                for klass in self.get_supplementary_klasses()
            ]
        )
        return " ".join(sources)

    def get_test_includes(self) -> List[str]:
        """
        Get the include statements for the test file.

        Returns:
            List[str]: A list of include statements for each class in the test file.
        """
        return [f'#include "{klass.to_snake_case()}.hpp"' for klass in self.classes]

    def get_supplementary_klasses(self):
        """
        Get the supplementary sources for the schema.

        Returns:
            List[str]: The supplementary sources.
        """
        return ["identifiable", "index"]

    def get_root_klass(self) -> "Klass":
        """
        Get the root klass for the schema.

        Returns:
            Klass: The root klass.

        Raises:
            ValueError: If no root klass is found.
        """
        for klass in self.classes:
            if not klass.has_parent():
                return klass
        raise ValueError("No root klass found")


@dataclass
class Klass:
    """
    A class used to define the structure of the data in a model.

    Attributes:
        name (str): The name of the class.

        description (str): The description of the class.

        fields (List[Field]): The fields in the class.

    """

    name: str
    description: str
    fields: List["Field"] = field(default_factory=list)
    _schema: Optional[Schema] = field(default=None, repr=False, init=False)

    def get_include_define(self) -> str:
        """
        Get the include define for the class.

        Returns:
            str: The include define.
        """
        return f"_{self.name.upper()}_HPP_"

    def get_forward_declarations(self) -> List[str]:
        """
        Get the forward declarations for the class.

        Returns:
            List[str]: The forward declarations.
        """
        forward_declarations = set()
        for field in self.fields:
            if field.is_reference():
                forward_declarations.add(field.type)
        return sorted(forward_declarations)

    def get_forward_includes(self) -> List[str]:
        """
        Get the forward includes for the class.

        Returns:
            List[str]: The forward includes.
        """
        return [to_snake_case(dec) for dec in self.get_forward_declarations()]

    def to_snake_case(self) -> str:
        """
        Convert the class name to snake_case.

        Returns:
            str: The class name in snake_case.
        """
        return to_snake_case(self.name)

    def init_fields(self, parents=True) -> List["Field"]:
        """
        Initialize the fields for the class.

        Args:
            parents (bool): Whether to include parent fields.

        Returns:
            List[Field]: The initialized fields.
        """
        init_fields = []
        for field in self.fields:
            if field.has_parent() and not parents:
                continue
            if not field.has_default() and not field.is_child and not field.is_optional:
                init_fields.append(field)
        return init_fields

    def get_create_arguments(self) -> str:
        """
        Get the create arguments for the class.

        Returns:
            str: The create arguments.
        """
        arguments = []
        for field in self.init_fields():
            arguments.append(f"{field.get_cpp_type()} {field.to_camel_case()}")
        return ", ".join(arguments)

    def get_example_arguments(self) -> str:
        """
        Get the example arguments for the class.

        Returns:
            str: The example arguments.
        """
        arguments = []
        for field in self.init_fields():
            arguments.append(field.get_example())
        return ", ".join(arguments)

    def get_example_update_arguments(self) -> str:
        """
        Get the example update arguments for the class.

        Returns:
            str: The example update arguments.
        """
        arguments = []
        for field in self.get_updatable_fields():
            if field.get_example() != "nullptr" and not field.has_parent():
                tuple_arg = f'{{"{field.to_camel_case()}", {field.get_example()}}}'
                arguments.append(tuple_arg)
        argument_str = ", ".join(arguments)

        return f"{{ {argument_str} }}"

    def get_ordered_fields(self):
        """
        Get the ordered fields for the class.

        Returns:
            List[Field]: The ordered fields.
        """
        # Returns a list of fields with the parent fields first, then other ordered alphabetically by name
        parent_fields = []
        other_fields = []
        for field in self.fields:
            if field.has_parent():
                parent_fields.append(field)
            else:
                other_fields.append(field)
        return sorted(parent_fields, key=lambda x: x.name) + sorted(
            other_fields, key=lambda x: x.name
        )

    def get_updatable_fields(self):
        """
        Get the updatable fields for the class.

        Returns:
            List[Field]: The updatable fields.
        """
        return [field for field in self.get_ordered_fields() if not field.is_list]

    def link(self) -> None:
        """
        Link the class to its fields and schema.
        """
        if not self._schema:
            raise ValueError("Schema is not linked")
        for field in self.fields:
            field._klass = self
            if field.has_parent():
                field._parent_klass = self._schema.get_klass(field.type)
                field._parent_field = self._schema.get_field(field.type, field.parent)  # type: ignore
            elif field.is_child:
                # Look for opposite field in child klass that has a matching parent
                child_klass = self._schema.get_klass(field.type)
                for child_field in child_klass.fields:
                    if child_field.has_parent() and child_field.type == self.name:
                        field._child_klass = child_klass
                        field._child_field = child_field
                        break

    def get_var_name(self) -> str:
        """
        Get the variable name for the class.

        Returns:
            str: The variable name.
        """
        # Return name with first letter lowercase
        return self.name[0].lower() + self.name[1:]

    def has_parent(self) -> bool:
        """
        Check if the class has a parent.

        Returns:
            bool: True if the class has a parent, False otherwise.
        """
        for field in self.fields:
            if field.has_parent():
                return True
        return False

    def get_create_ptr_type(self) -> str:
        """
        Get the create pointer type for the class.

        Returns:
            str: The create pointer type.
        """
        if self.has_parent():
            return f"std::weak_ptr<{self.name}>"
        else:
            return f"std::shared_ptr<{self.name}>"

    def get_parent_fields(self) -> List["Field"]:
        """
        Get the parent fields for the class.

        Returns:
            List[Field]: The parent fields.
        """
        fields = []
        for field in self.fields:
            if field.has_parent():
                if field._parent_klass is not None:
                    fields.extend(field._parent_klass.get_parent_fields())
                fields.append(field)
        return fields

    def get_parent_field(self) -> "Field":
        """
        Get the parent field for the class.

        Returns:
            Field: The parent field.

        Raises:
            ValueError: If no parent field is found.
        """
        for field in self.fields:
            if field.has_parent():
                return field
        raise ValueError("No parent field found")

    def is_root(self) -> bool:
        """
        Check if the class is the root klass.

        Returns:
            bool: True if the class is the root klass, False otherwise.
        """
        return not self.has_parent()


@dataclass
class Field:
    """
    A field used to define the structure of the data in a model.

    Attributes:
        name (str): The name of the field.

        description (str): The description of the field.

        type (str): The type of the field.

        example (Any): The example value of the field.

        default (Optional[Any]): The default value of the field.

        parent (Optional[str]): The parent of the field.

        is_child (bool): Whether the field is a child.

        is_list (bool): Whether the field is a list.

        is_optional (bool): Whether the field is optional.

    """

    name: str
    description: str
    type: str
    example: Optional[Any] = None
    default: Optional[Any] = None
    parent: Optional[str] = None
    is_child: bool = False
    is_list: bool = False
    is_optional: bool = False
    _parent_klass: Optional[Klass] = field(default=None, repr=False, init=False)
    _parent_field: Optional["Field"] = field(default=None, repr=False, init=False)
    _child_klass: Optional[Klass] = field(default=None, repr=False, init=False)
    _child_field: Optional["Field"] = field(default=None, repr=False, init=False)
    _klass: Optional[Klass] = field(default=None, repr=False, init=False)

    def has_parent(self) -> bool:
        """
        Check if the field has a parent.

        Returns:
            bool: True if the field has a parent, False otherwise.
        """
        return self.parent is not None

    def has_default(self) -> bool:
        """
        Check if the field has a default value.

        Returns:
            bool: True if the field has a default value, False otherwise.
        """
        return self.default is not None

    def to_camel_case(self, upper_first=False) -> str | None:
        """
        Convert the field name to camelCase.

        Args:
            upper_first (bool): Whether to capitalize the first letter.

        Returns:
            str: The field name in camelCase.
        """
        return to_camel_case(self.name, upper_first)

    def get_cpp_type(self, nolist=False, cast=False) -> str:
        """
        Get the C++ type for the field.

        Args:
            nolist (bool): Whether to exclude list types.
            cast (bool): Whether to cast the type.

        Returns:
            str: The C++ type.
        """
        # Handle parent fields
        if self.has_parent():
            if cast:
                return f"std::shared_ptr<{self.type}>"
            else:
                return f"std::weak_ptr<{self.type}>"

        # Handle child fields
        if self.is_child:
            if self.is_list and not nolist:
                return f"std::vector<std::shared_ptr<{self.type}>>"
            else:
                return f"std::shared_ptr<{self.type}>"

        # Handle reference fields
        if self.is_reference():
            if self.is_list and not nolist:
                return f"std::vector<std::weak_ptr<{self.type}>>"
            else:
                return f"std::weak_ptr<{self.type}>"

        # Handle other fields
        if self.type in TYPEMAP:
            type = TYPEMAP[self.type][0]
        else:
            type = self.type

        # Handle optional fields
        if self.is_optional:
            type = f"std::optional<{type}>"

        return type

    def is_reference(self) -> bool:
        """
        Check if the field is a reference.

        Returns:
            bool: True if the field is a reference, False otherwise.

        Raises:
            ValueError: If the class or schema is not linked.
        """
        if self._klass is None:
            raise ValueError("Klass is not linked")

        if self._klass._schema is None:
            raise ValueError("Schema is not linked")

        return self.type in [klass.name for klass in self._klass._schema.classes]

    def get_example(self) -> Any:
        """
        Get the example value for the field.

        Returns:
            Any: The example value.
        """
        example = self.example
        if self._parent_field is not None:
            example = self.to_camel_case()
        elif isinstance(example, str):
            example = f'std::string("{example}")'
        elif isinstance(example, list):
            example = "{}"
        elif example is None:
            example = "nullptr"
        elif isinstance(example, bool):
            example = str(example).lower()
        elif isinstance(example, float):
            example = f"{example}f"
        return example



TEMPLATE = """

#ifndef {{klass.get_include_define()}}
#define {{klass.get_include_define()}}

#include <any>
#include <map>
#include <memory>
#include <optional>
#include <string>
#include <vector>
#include <fstream>

#include "identifiable.hpp"
#include "persistable.hpp"

namespace {{schema.namespace}}
{
{%- for forward in klass.get_forward_declarations() %}
    class {{forward}};
{%- endfor %}


    /**
        @brief {{klass.description}}

{%- for field in klass.get_ordered_fields() %}
        @param {{field.to_camel_case()}} ({{field.get_cpp_type()}}) {{field.description}}.{% if field.has_default() %} Defaults to {{field.default}}.{% endif %}
{%- endfor %}
    */

    class {{klass.name}} : public std::enable_shared_from_this<{{klass.name}}>, public Identifiable, public Persistable<{{klass.name}}>
    {
    private:
        struct PrivateConstructor
        {
            PrivateConstructor() = default;
        };

{%- for field in klass.get_ordered_fields() %}
        {{field.get_cpp_type()}} {{field.to_camel_case()}}{% if field.has_default() %} = {{field.default}}{% endif %};
{%- endfor %}

    public:
        {{klass.name}}(PrivateConstructor) {}


        /**
            @brief Create a new {{klass.name}} object
{%- for field in klass.init_fields() %}
            @param {{field.to_camel_case()}} {{field.description}}
{%- endfor %}
            @return A pointer to the new object
        */
        static {{klass.get_create_ptr_type()}} create({{klass.get_create_arguments()}});

        /**
            @brief Update the object with new values
            @param fields A map of field names to new values
        */
        void update(std::map<std::string, std::any> fields);

        /**
            @brief Destroy the object and remove it from parent objects, also removes all child objects
        */
        void destroy(bool fromParent = false);

        {{klass.name}}() : {{klass.name}}(PrivateConstructor()) {}

        /**
            @brief Get a shared pointer to the object, use this instead of make_shared
            @return A shared pointer to the object
        */
        std::shared_ptr<{{klass.name}}> getptr();
        
{%- for field in klass.get_ordered_fields() %}
    {%- if field.has_parent(): %}
        /**
            @brief Set this object for
        */
        void reset{{field.to_camel_case(upper_first=True)}}();
    {%- endif %}
    {% if field.is_list %}
        {%- if field.child_field %}
        /**
            @brief Add an item to the list of {{field.to_camel_case()}}
            @param item The item to add
            @param fromChild True if this is called from the child object
        */
        void addTo{{field.to_camel_case(upper_first=True)}}({{field.get_cpp_type(nolist=True)}} item, bool fromChild = false);

        /**
            @brief Remove an item from the list of {{field.to_camel_case()}}
            @param item The item to remove
            @param fromChild True if this is called from the child object
        */
        void removeFrom{{field.to_camel_case(upper_first=True)}}({{field.get_cpp_type(nolist=True)}} item, bool fromChild = false);
        {%- else %}

        /**  
            @brief Add an item to the list of {{field.to_camel_case()}}
            @param item The item to add
        */
        void addTo{{field.to_camel_case(upper_first=True)}}({{field.get_cpp_type(nolist=True)}} item, bool fromChild = false);
        
        /**
            @brief Remove an item from the list of {{field.to_camel_case()}}
            @param item The item to remove
        */
        void removeFrom{{field.to_camel_case(upper_first=True)}}({{field.get_cpp_type(nolist=True)}} item, bool fromChild = false);
        {%- endif %}
    {% else %}
        /**
            @brief Set the value of {{field.to_camel_case()}}
            @param {{field.to_camel_case()}} The new value
        */
        void set{{field.to_camel_case(upper_first=True)}}({{field.get_cpp_type()}} {{field.to_camel_case()}});
    {% endif %}
    
        /**
            @brief Get the value of {{field.to_camel_case()}}
            @return The value of {{field.to_camel_case()}}
        */
        {{field.get_cpp_type()}} get{{field.to_camel_case(upper_first=True)}}();

{%- endfor %}

        /**
         * @brief Add the object to an index
         * @param index The index to add the object to
         */
        void addToIndex(std::shared_ptr<Index> index);
        
        /**
         * @brief Serialize the object's children to a stream
         *@param os The stream to serialize to
         */
        void serializeFields(std::ostream &os);

        /**
         * @brief Serialize the object's references to a stream
         * @param os The stream to serialize to
         */
        void serializeReferences(std::ostream &os);

        /**
         * @brief Deserialize the object's children from a stream
         * @param is The stream to deserialize from
         * @param index The index to use for reference resolution
         */
        static std::shared_ptr<{{klass.name}}> deserializeFields(std::istream &is, std::shared_ptr<Index> index);

        /**
         * @brief Deserialize the object's references from a stream
         * @param is The stream to deserialize from
         * @param index The index to use for reference resolution
         */
        void deserializeReferences(std::istream &is, std::shared_ptr<Index> index);
    

    };
}
#endif
"""
